{*******************************************************}
{                                                       }
{ Copyright(C) Linus Gates & Bill Torvalds 2003-2018    }
{                                                       }
{*******************************************************}
{}
{  -     BMP   }
{      3.90                    }
{}

{$ifdef fpc}
{$ASMMODE intel}
{$A1}
{$endif}

{$R-}

Unit Bmpunit;

Interface

Type
  PScrArray = ^TScrArray;
  TScrArray = Array[0..0] Of LongInt;

Const
  bi_RGB  = 0;
  bi_RLE8 = 1;
  bi_RLE4 = 2;

Type
  TRGBQuad = Record
    rgbBlue: Byte;
    rgbGreen: Byte;
    rgbRed: Byte;
    rgbReserved: Byte;
  End;

Type
  PBitmapInfoHeader = ^TBitmapInfoHeader;
  TBitmapInfoHeader = Record
    biSize: LongInt;
    biWidth: LongInt;
    biHeight: LongInt;
    biPlanes: Word;
    biBitCount: Word;
    biCompression: LongInt;
    biSizeImage: LongInt;
    biXPelsPerMeter: LongInt;
    biYPelsPerMeter: LongInt;
    biClrUsed: LongInt;
    biClrImportant: LongInt;
  End;

Type
  PBitmapInfo = ^TBitmapInfo;
  TBitmapInfo = Record
    bmiHeader: TBitmapInfoHeader;
{   Bmicolors: Array[0..0] Of TRGBquad; }
  End;

Type
  PBitmapFileHeader = ^TBitmapFileHeader;
  TBitmapFileHeader = Record
    bfType: Word;
    bfSize: LongInt;
    bfReserved1: Word;
    bfReserved2: Word;
    bfOffBits: LongInt;
  End;

Type
  PBitmap = ^TBitmap;
  TBitmap = Object
    Xx,Yy:Word;
    Nposx,Nposy:Word;
    ScrArray:PScrArray;
    F:File;
    Error:Word;
    MaxColor:DWord;
    Constructor Init(FileName:String);
    Procedure ShowBitmap(X,Y:Word);Virtual;
    Procedure Draw;Virtual;
    Destructor Done;
  End;

function _32to16bit(Color:LongInt):Word;

Implementation

Uses Graph;

function _32to16bit(Color:LongInt):Word;
Var
  RGBColor:TRGBQuad absolute Color;
begin
  RGBColor.rgbRed:=   RGBColor.rgbRed   shr 3;
  RGBColor.rgbGreen:= RGBColor.rgbGreen shr 2;
  RGBColor.rgbBlue:=  RGBColor.rgbBlue  shr 3;
  _32to16bit:=Word(RGBColor.rgbRed shl 11) or Word(RGBColor.rgbGreen shl 5) or Word(RGBColor.rgbBlue);
end;

Function R32(Ddx:LongInt):LongInt;
begin
  asm
    mov eax,Ddx
    and eax,$FFFFFFFC
    cmp eax,Ddx
    je @1
    add eax,4
@1: mov Ddx,eax
  end;
  R32:=Ddx;
end;

Constructor TBitmap.Init(FileName: String);
Var
  Fh:TBitmapFileHeader;
  Bi:TBitmapInfo;
  K:Byte;
  Massiv:Array[0..5000] Of Byte;
  DX:LongInt;
  DY:LongInt;
Procedure Load24Bitmap;
Var
   J:LongInt;
   I:LongInt;
Begin
  If Error <> 0 Then
     Exit;
  Seek(F,Fh.bfOffBits);
  DX:=DX*3;
  DX:=R32(DX);
  For J:=Dy-1 Downto 0 Do
  Begin
    BlockRead(F,Massiv,DX);
    I:=0;
    While I<Bi.bmiHeader.biWidth*3 Do
    Begin
      ScrArray^[LongInt(Bi.bmiHeader.biWidth) * J + (I Div 3)]:=Massiv[I]+(Massiv[I+1]ShL 8)+(Massiv[I+2] ShL 16);
      Inc(I,3);
    End;
  End;
End;
Procedure Load256Bitmap;
Var
  ColorArray:Array[0..255] Of LongInt;
  RGB:TRGBQuad;
  I:LongInt;
  J:LongInt;
Begin
  If Error<>0 Then Exit;
  DX:=R32(DX);
  For I:=0 To 255 Do
  Begin
    BlockRead(F,RGB,SizeOf(RGB));
    ColorArray[I]:=RGB.rgbRed ShL 16 + RGB.rgbGreen ShL 8 + RGB.rgbBlue;
  End;
  Seek(F,Fh.bfOffBits);
  For J:=Dy-1 Downto 0 Do
  Begin
    BlockRead(F,Massiv,DX);
    For I:=0 To Bi.bmiHeader.biWidth-1 Do
      ScrArray^[LongInt(Bi.bmiHeader.biWidth)*J+I]:=ColorArray[Massiv[I]];
  End;
End;
Procedure Load16Bitmap;
Var
  ColorArray:Array[0..15] Of LongInt;
  RGB:TRGBQuad;
  I:LongInt;
  J:LongInt;
Begin
  If Error <> 0 Then
    Exit;
  For I:=0 To 15 Do
  Begin
    BlockRead(F,RGB,SizeOf(RGB));
    ColorArray[I]:=RGB.rgbRed ShL 16 + RGB.rgbGreen ShL 8 + RGB.rgbBlue;
  End;
  Seek(F,Fh.bfOffBits);
  DX:=Bi.bmiHeader.biWidth;
  If (DX and 1) = 1 Then
    Inc(DX);
  DX:=DX shr 1;
  DX:=R32(DX);
  For J:=Dy-1 Downto 0 Do
  Begin
    BlockRead(F, Massiv, DX);
    For I:=0 to Bi.bmiHeader.biWidth-1 Do
      Begin
        K:=Massiv[I Div 2];
        If I and $1 = 0 Then
          ScrArray^[LongInt(Bi.bmiHeader.biWidth)*J+I]:=ColorArray[K shr 4]
        Else
          ScrArray^[LongInt(Bi.bmiHeader.biWidth)*J+I]:=ColorArray[K and $0F];
      End;
  End;
End;
Procedure LoadMonoBitmap;
Var
  ColorArray:Array[0..1] Of LongInt;
  RGB:TRGBQuad;
  J:LongInt;
  I:LongInt;
Begin
  If Error<>0 Then Exit;
  BlockRead(F,RGB,SizeOf(RGB));
  ColorArray[0]:=RGB.rgbRed ShL 16 + RGB.rgbGreen ShL 8 + RGB.rgbBlue;
  BlockRead(F,RGB,SizeOf(RGB));
  ColorArray[1]:=RGB.rgbRed ShL 16 + RGB.rgbGreen ShL 8 + RGB.rgbBlue;
  Seek(F,Fh.bfOffBits);
  DX:=Bi.bmiHeader.biWidth+7;
  DX:=DX div 8;
  DX:=R32(DX);
  For J:=Bi.bmiHeader.biHeight-1 Downto 0 Do
  Begin
    I:=0;
    BlockRead(F,Massiv,DX);
    For I:=0 to Bi.bmiHeader.biWidth-1 Do
    begin
      K:=Massiv[I Div 8];
      If (K and ($80 shr (I mod 8)))=0 Then
        ScrArray^[LongInt(Bi.bmiHeader.biWidth)*J+I]:=ColorArray[0] Else ScrArray^[LongInt(Bi.bmiHeader.biWidth)*J+I]:=ColorArray[1];
    end;
  End;
End;
Procedure Load32Bitmap;
Var
  J:LongInt;
  I:LongInt;
Begin
  If Error<>0 Then Exit;
  Seek(F,Fh.bfOffBits);
  DX:=DX*4;
  DX:=R32(DX);
  For J:=Dy-1 Downto 0 Do
  Begin
    BlockRead(F,Massiv,DX);
    I:=0;
    While I<Bi.bmiHeader.biWidth*4 Do
    Begin
      ScrArray^[LongInt(Bi.bmiHeader.biWidth)*J+(I Div 4)]:=Massiv[I]+(Massiv[I+1]ShL 8)+(Massiv[I+2] ShL 16);
      Inc(I,4);
    End;
  End;
End;

Begin
  Error:=0;
  ScrArray:=Nil;
  Assign(F,FileName);
  {$I-}
  Reset(F,1);
  {$I+}
  If IOResult <> 0 Then
  Begin
    Error:=1;
    Exit;
  End;
  BlockRead(F,Fh,SizeOf(TBitmapFileHeader));
  BlockRead(F,Bi,SizeOf(TBitmapInfo));
  If Bi.bmiHeader.biCompression<>bi_RGB Then Error:=2;
  Dy:=Bi.bmiHeader.biHeight;
  DX:=Bi.bmiHeader.biWidth;
  GetMem(ScrArray, LongInt(Bi.bmiHeader.biHeight) * LongInt(Bi.bmiHeader.biWidth) * SizeOf(LongInt));
  Case Bi.bmiHeader.biBitCount Of
    1:LoadMonoBitmap;
    4:Load16Bitmap;
    8:Load256Bitmap;
    24:Load24Bitmap;
    32:Load32Bitmap;
    Else Error:=3;
  End;
  Close(F);
  Xx:=Bi.bmiHeader.biWidth-1;
  Yy:=Bi.bmiHeader.biHeight-1;
End;

Procedure TBitmap.Draw;
Var
  I,J:Word;
Begin
  MaxColor:=GetMaxColor;
  For J:=0 To Yy{-1} Do
    For I:=0 To Xx{-1} Do
    Case MaxColor Of
    $FFFF:
      PutPixel(I+Nposx,J+Nposy,_32to16bit(ScrArray^[LongInt(XX+1)*(J)+I]));
    else
      PutPixel(I+Nposx,J+Nposy,ScrArray^[LongInt(XX+1)*(J)+I]);
    end;
End;

Procedure TBitmap.ShowBitmap(X,Y:Word);
Begin
  Nposx:=X;
  Nposy:=Y;
  Draw;
End;

Destructor TBitmap.Done;
Begin
  If Assigned(ScrArray) Then
    FreeMem(ScrArray, LongInt(Xx + 1) * LongInt(Yy + 1) * SizeOf(LongInt));
End;

End.
